from minibert.attention import AttentionType
from .embeddings import PositionalEmbeddingType

__all__ = [
    "MiniBertConfiguration",
    "MiniBertForMLMConfiguration",
    "MiniBertForRegressionConfiguration",
    "MiniBertForMLMAndRegressionConfiguration",
]


class MiniBertConfiguration:
    def __init__(self, **kwargs):
        self.vocabulary = kwargs["vocabulary"]

        # Attention configuration
        self.embedding_dim = kwargs.get("embedding_dim", 64)
        self.hidden_dim = kwargs.get("hidden_dim", self.embedding_dim)
        self.key_is_query = kwargs.get("key_is_query", False)
        self.attention_type = kwargs.get(
            "attention_type", AttentionType.SelfAttention)

        # Embeddings configuration
        self.position_embeddings_count = kwargs.get(
            "position_embeddings_count", 1024)
        self.position_type = kwargs.get(
            "position_type", PositionalEmbeddingType.TRAINED)
        self.normalize_embeddings = kwargs.get(
            "normalize_embeddings", True)


class MiniBertForMLMConfiguration(MiniBertConfiguration):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.mask_idx = kwargs["mask_idx"]
        self.mask_token = kwargs.get("mask_token", "<mask>")

        # Masking strategy
        self.mask_prob = kwargs.get("mask_prob", 0.15)
        self.keep_mask_prob = kwargs.get("keep_mask_prob", 0.8)
        self.corrupt_mask_prob = kwargs.get("corrupt_mask_prob", 0.1)
        self.reveal_mask_prob = kwargs.get("reveal_mask_prob", 0.1)

        if self.keep_mask_prob + self.corrupt_mask_prob + self.reveal_mask_prob != 1:
            raise ValueError("Sum of masking strategies is not 1")

        # Prediction layers
        self.first_layer_output_size = kwargs.get(
            "first_layer_output_size", self.embedding_dim)
        self.first_layer_activation_fun = kwargs.get("activation_fun", "gelu")
        self.second_layer_activation_fun = kwargs.get("activation_fun", "none")


class MiniBertForRegressionConfiguration(MiniBertConfiguration):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Prediction layers
        self.first_layer_output_size = kwargs.get(
            "first_layer_output_size", self.embedding_dim)
        self.first_layer_activation_fun = kwargs.get("activation_fun", "gelu")
        self.second_layer_activation_fun = kwargs.get("activation_fun", "none")
        self.output_size = kwargs["output_size"]


class MiniBertForMLMAndRegressionConfiguration(MiniBertConfiguration):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.mask_idx = kwargs["mask_idx"]
        self.mask_token = kwargs.get("mask_token", "<mask>")

        # Masking strategy
        self.mask_prob = kwargs.get("mask_prob", 0.15)
        self.keep_mask_prob = kwargs.get("keep_mask_prob", 0.8)
        self.corrupt_mask_prob = kwargs.get("corrupt_mask_prob", 0.1)
        self.reveal_mask_prob = kwargs.get("reveal_mask_prob", 0.1)

        if self.keep_mask_prob + self.corrupt_mask_prob + self.reveal_mask_prob != 1:
            raise ValueError("Sum of masking strategies is not 1")

        # Prediction layers

        # MLM
        self.mlm_first_layer_output_size = kwargs.get(
            "mlm_first_layer_output_size", self.embedding_dim)
        self.mlm_first_layer_activation_fun = kwargs.get(
            "activation_fun", "gelu")
        self.mlm_second_layer_activation_fun = kwargs.get(
            "activation_fun", "none")

        # Regression
        self.reg_first_layer_output_size = kwargs.get(
            "reg_first_layer_output_size", self.embedding_dim)
        self.reg_first_layer_activation_fun = kwargs.get(
            "activation_fun", "gelu")
        self.reg_second_layer_activation_fun = kwargs.get(
            "activation_fun", "none")
        self.reg_output_size = kwargs["reg_output_size"]



from enum import Enum
from math import sqrt

from .embeddings import PositionalEmbeddingType, PositionalEmbedding

import torch
from torch import nn
from torch.nn import functional as F

__all__ = [
    "AttentionType",
    "AttentionEmbedding",
    "Attention",
    "NonTransformingAttention",
]


class AttentionType(Enum):
    SelfAttention = 1
    AttentionEmbedding = 2
    NonTransformingAttention = 3


class AttentionEmbedding(nn.Module):
    def __init__(self, embedding_dim, voc_size, out_dim=None, max_seq_len=1024, position_type=PositionalEmbeddingType.TRAINED, normalize_embeddings=True):
        super(AttentionEmbedding, self).__init__()
        if out_dim is None:
            out_dim = embedding_dim

        self.embedding_dim = embedding_dim
        self.voc_size = voc_size
        self.out_dim = out_dim
        self.max_seq_len = max_seq_len

        self.key = nn.Embedding(voc_size, embedding_dim)
        self.query = nn.Embedding(voc_size, embedding_dim)
        self.value = nn.Embedding(voc_size, out_dim)
        self._sqrt_embedding = sqrt(embedding_dim)

        self.position_embedding = PositionalEmbedding(
            embedding_dim, max_seq_len, ptype=position_type)

        self.norm = None
        self.normalize_embeddings = normalize_embeddings
        if normalize_embeddings:
            self.norm = nn.LayerNorm(embedding_dim)

    def forward(self, input):
        pos = self.position_embedding(input)
        key = self.key(input) + pos
        query = self.query(input) + pos
        value = self.value(input)

        if self.normalize_embeddings:
            key = self.norm(key)
            query = self.norm(query)
            value = self.norm(value)

        key_t = torch.transpose(key, -2, -1)
        qk = torch.matmul(query, key_t) / self._sqrt_embedding
        attention = F.softmax(qk, dim=-1)
        return torch.matmul(attention, value)


class Attention(nn.Module):
    def __init__(self, in_dim, out_dim, hidden_dim=None, key_is_query=False):
        super(Attention, self).__init__()
        if hidden_dim is None:
            hidden_dim = out_dim

        self.in_dim = in_dim
        self.out_dim = out_dim
        self.key_is_query = key_is_query
        self.hidden_dim = hidden_dim
        self.key = nn.Parameter(torch.rand((in_dim, hidden_dim)))
        if key_is_query:
            self.query = self.key
        else:
            self.query = nn.Parameter(torch.rand((in_dim, hidden_dim)))
        self.value = nn.Parameter(torch.rand((in_dim, out_dim)))
        self._sqrt_hidden = sqrt(hidden_dim)

    @classmethod
    def from_weights(cls, key, query, value):
        in_dim, hidden_dim = key.shape
        out_dim = value.shape[1]
        x = cls(in_dim, out_dim, hidden_dim)
        with torch.no_grad():
            x.key = nn.Parameter(key)
            x.query = nn.Parameter(query)
            x.value = nn.Parameter(value)
        return x

    def forward(self, input):
        key = torch.matmul(input, self.key)
        query = torch.matmul(input, self.query)
        value = torch.matmul(input, self.value)

        key_t = torch.transpose(key, -2, -1)
        qk = torch.matmul(query, key_t) / self._sqrt_hidden
        attention = F.softmax(qk, dim=-1)
        return torch.matmul(attention, value)


class NonTransformingAttention(nn.Module):
    def __init__(self, dim):
        super(NonTransformingAttention, self).__init__()
        self.dim = dim
        self._sqrt_dim = sqrt(dim)

    def forward(self, input):
        query = input
        key = input
        key_t = torch.transpose(key, -2, -1)
        qk = torch.matmul(query, key_t) / self._sqrt_dim
        attention = F.softmax(qk, dim=-1)
        return torch.matmul(attention, input)



import torch.nn.functional as F


def parse_activation_function(act):
    act = act.strip().lower()
    if act in ("nothing", "none"):
        return lambda x: x
    else:
        return getattr(F, act)



import unittest
import torch
from torch.nn import functional as F
from math import sqrt

from minibert import Attention


class TestAttention(unittest.TestCase):
    def test_attention_given_matrix(self):
        k = torch.tensor([[0, 0.5], [1, 0], [0.5, 0.5]], dtype=torch.float)
        q = torch.tensor([[0, 0.5], [0, 0], [0.5, 0.5]], dtype=torch.float)
        v = torch.tensor([[0.5, 0.5], [1, 0.5], [1, 1]], dtype=torch.float)
        attention = Attention.from_weights(k, q, v)

        x = torch.tensor(
            [[1, 0, 1], [1, 1, 1], [0, 0, 1]], dtype=torch.float)

        xk = torch.tensor([[0.5, 1], [1.5, 1], [0.5, 0.5]], dtype=torch.float)
        xq = torch.tensor([[0.5, 1], [0.5, 1], [0.5, 0.5]], dtype=torch.float)
        xv = torch.tensor([[1.5, 1.5], [2.5, 2], [1, 1]], dtype=torch.float)

        x_qk = torch.matmul(xq, xk.t()) / sqrt(2)
        expected = torch.matmul(F.softmax(x_qk, dim=1), xv)

        actual = attention(x)
        self.assertTrue(torch.equal(expected, actual))

    def test_attention_given_batch(self):
        k = torch.tensor([[0, 0.5], [1, 0], [0.5, 0.5]], dtype=torch.float)
        q = torch.tensor([[0, 0.5], [0, 0], [0.5, 0.5]], dtype=torch.float)
        v = torch.tensor([[0.5, 0.5], [1, 0.5], [1, 1]], dtype=torch.float)
        attention = Attention.from_weights(k, q, v)

        x = torch.tensor(
            [[1, 0, 1], [1, 1, 1], [0, 0, 1]], dtype=torch.float)
        batch = torch.stack([x, x, x])

        xk = torch.tensor([[0.5, 1], [1.5, 1], [0.5, 0.5]], dtype=torch.float)
        xq = torch.tensor([[0.5, 1], [0.5, 1], [0.5, 0.5]], dtype=torch.float)
        xv = torch.tensor([[1.5, 1.5], [2.5, 2], [1, 1]], dtype=torch.float)

        x_qk = torch.matmul(xq, xk.t()) / sqrt(2)
        expected = torch.matmul(F.softmax(x_qk, dim=1), xv)
        expected = torch.stack([expected, expected, expected])

        actual = attention(batch)
        self.assertTrue(torch.equal(expected, actual))


if __name__ == '__main__':
    unittest.main()



from minibert.embeddings import PositionalEmbeddingType
import unittest
import torch

from minibert import *


class TestMiniBert(unittest.TestCase):
    def test_minibert_not_fail(self):
        vocabulary = ["<mask>", "a", "b", "c", "d", "e"]
        d = 4

        for pos in PositionalEmbeddingType:
            for att in AttentionType:
                config = MiniBertConfiguration(
                    vocabulary=vocabulary,
                    embedding_dim=d,
                    position_embeddings_count=16,
                    mask_idx=0,
                    position_type=pos,
                    attention_type=att
                )
                minibert = MiniBert(config)
                x = torch.tensor([
                    [1, 3, 4, 5],
                    [1, 3, 4, 5],
                    [1, 3, 4, 5]
                ])
                output = minibert(x)
                self.assertEqual(
                    output.size(),
                    torch.Size((x.size(0), x.size(1), d))
                )

    def test_minibert_mlm_and_reggression(self):
        vocabulary = ["<mask>", "a", "b", "c", "d", "e"]
        sentences = torch.tensor([
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
        ])

        x_acr = torch.tensor([
            [3, 4, 5],
            [1, 4, 2],
        ])
        y_acr = torch.tensor([1, 2])

        for pos in PositionalEmbeddingType:
            for att in AttentionType:
                config = MiniBertForMLMAndRegressionConfiguration(
                    vocabulary=vocabulary,
                    embedding_dim=4,
                    position_embeddings_count=16,
                    mask_idx=0,
                    reg_output_size=len(vocabulary),
                    position_type=pos,
                    attention_type=att
                )
                model = MiniBertForMLMAndRegression(config)

                mlm_output = model.forward(sentences, 0)
                reg_output = model.forward(x_acr, 1)

                self.assertIsInstance(mlm_output, tuple)
                self.assertEqual(
                    mlm_output[0].size(),
                    torch.Size(
                        (sentences.size(0), sentences.size(1), len(vocabulary)))
                )
                self.assertEqual(
                    reg_output.size(),
                    torch.Size((x_acr.size(0), x_acr.size(1), len(vocabulary)))
                )


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from math import sin, cos

import torch
from torch import nn

__all__ = [
    "PositionalEmbeddingType",
    "PositionalEmbedding",
    "MiniBertEmbedding",
]


class PositionalEmbeddingType(Enum):
    TRAINED = 1
    FIXED = 2
    NONE = 3


class PositionalEmbedding(nn.Module):
    def __init__(self, embedding_dim, max_seq_len, ptype=PositionalEmbeddingType.TRAINED):
        super().__init__()

        self.embedding_dim = embedding_dim
        self.max_seq_len = max_seq_len

        if ptype == PositionalEmbeddingType.TRAINED:
            self.embeddings = nn.Embedding(max_seq_len, embedding_dim)
        elif ptype == PositionalEmbeddingType.FIXED:
            # See Attention is all you need, section 3.5 (https://arxiv.org/pdf/1706.03762.pdf)
            positions = torch.zeros(
                (max_seq_len, embedding_dim), dtype=torch.float)
            for pos in range(max_seq_len):
                for i in range(embedding_dim):
                    if i % 2 == 0:
                        positions[pos, i] = sin(
                            pos / pow(10000, 2 * i / embedding_dim))
                    else:
                        positions[pos, i] = cos(
                            pos / pow(10000, 2 * i / embedding_dim))
            self.embeddings = nn.Embedding.from_pretrained(
                positions, freeze=True)
        elif ptype == PositionalEmbeddingType.NONE:
            positions = torch.zeros(
                (max_seq_len, embedding_dim), dtype=torch.float)
            self.embeddings = nn.Embedding.from_pretrained(
                positions, freeze=True)
        else:
            raise Exception("Invalid position type")

        self.register_buffer(
            "position_ids", torch.arange(max_seq_len).expand((1, -1)))

    def forward(self, input):
        seq_len = input.shape[-1]
        return self.embeddings(self.position_ids[:, :seq_len])


class MiniBertEmbedding(nn.Module):
    def __init__(self, voc_size, embedding_dim, max_seq_len, position_type, normalize_embeddings):
        super().__init__()

        self.max_seq_len = max_seq_len
        self.position_type = position_type
        self.normalize_embeddings = normalize_embeddings

        self.word_embeddings = nn.Embedding(voc_size, embedding_dim)
        self.position_embeddings = PositionalEmbedding(
            embedding_dim, max_seq_len, ptype=position_type)

        self.norm = None
        if normalize_embeddings:
            self.norm = nn.LayerNorm(embedding_dim)

    def forward(self, input):
        word_emb = self.word_embeddings(input)
        pos_emb = self.position_embeddings(input)
        emb = word_emb + pos_emb

        if self.normalize_embeddings:
            return self.norm(emb)
        else:
            return emb



import setuptools

with open("README.md", "rt", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="minibert",
    version="0.2.1",
    author="Gaëtan Caillaut",
    author_email="gaetan.caillaut@univ-lemans.fr",
    description="A simplified implementation of BERT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git-lium.univ-lemans.fr/gcaillaut/minibert",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



import torch
from torch import nn
from torch.nn import functional as F
from math import sqrt, sin, cos, pow
from enum import Enum

from .activations import parse_activation_function
from .attention import Attention, AttentionType, AttentionEmbedding, NonTransformingAttention
from .embeddings import MiniBertEmbedding

__all__ = [
    "MiniBert",
    "MiniBertForMLM",
    "MiniBertForRegression",
    "MiniBertForMLMAndRegression",
]


class MiniBert(nn.Module):
    def __init__(self, configuration):
        super(MiniBert, self).__init__()

        self.configuration = configuration

        self._voc_size = len(configuration.vocabulary)
        self._embedding_dim = configuration.embedding_dim

        if configuration.attention_type == AttentionType.AttentionEmbedding:
            self.attention_embedding = AttentionEmbedding(
                self._embedding_dim,
                self._voc_size,
                position_type=configuration.position_type,
                normalize_embeddings=configuration.normalize_embeddings
            )
        else:
            self.embedding = MiniBertEmbedding(
                self._voc_size,
                self._embedding_dim,
                max_seq_len=configuration.position_embeddings_count,
                position_type=configuration.position_type,
                normalize_embeddings=configuration.normalize_embeddings
            )

            if configuration.attention_type == AttentionType.SelfAttention:
                self.attention = Attention(
                    self._embedding_dim,
                    self._embedding_dim,
                    hidden_dim=configuration.hidden_dim,
                    key_is_query=configuration.key_is_query
                )
            elif configuration.attention_type == AttentionType.NonTransformingAttention:
                self.attention = NonTransformingAttention(
                    self._embedding_dim
                )
            else:
                raise Exception("Invalid attention type")

    def forward(self, input):
        if self.configuration.attention_type == AttentionType.AttentionEmbedding:
            x = self.attention_embedding(input)
        else:
            x = self.embedding(input)
            x = self.attention(x)
        return x


class MiniBertForMLM(nn.Module):
    def __init__(self, configuration):
        super().__init__()
        self.minibert = MiniBert(configuration)
        self.configuration = configuration

        self._voc_size = len(configuration.vocabulary)
        self._embedding_dim = configuration.embedding_dim

        self.l1 = nn.Linear(self._embedding_dim,
                            configuration.first_layer_output_size, bias=False)
        self.l2 = nn.Linear(
            configuration.first_layer_output_size, self._voc_size, bias=True)

        self.mask_idx = configuration.mask_idx

        self.l1_activation_fun = parse_activation_function(
            configuration.first_layer_activation_fun)
        self.l2_activation_fun = parse_activation_function(
            configuration.second_layer_activation_fun)

        self._mask_prob = configuration.mask_prob
        self._keep_mask_prob = configuration.keep_mask_prob
        self._inv_corrupt_mask_prob = 1 - configuration.corrupt_mask_prob

    def forward(self, input):
        if self.training:
            masked_input = input.detach().clone()
            masked = torch.rand_like(
                input, dtype=torch.float) <= self._mask_prob

            masking_strategy = torch.rand_like(input, dtype=torch.float)
            masking = masked & (masking_strategy <=
                                self._keep_mask_prob)  # Keep masks
            corrupt = masked & (self._inv_corrupt_mask_prob <
                                masking_strategy)  # Corrupt masks

            replacements = torch.randint(
                self._voc_size, (torch.sum(corrupt), ), device=input.device)

            masked_input[masking] = self.mask_idx
            masked_input[corrupt] = replacements
            x = self.minibert(masked_input)
        else:
            x = self.minibert(input)

        x = self.l1(x)
        x = self.l1_activation_fun(x)
        x = self.l2(x)
        x = self.l2_activation_fun(x)

        if self.training:
            labels = input.detach().clone()
            labels[~masked] = -1

            loss_fn = nn.CrossEntropyLoss(ignore_index=-1)
            loss = loss_fn(x.view(-1, self._voc_size), labels.view(-1))
            return (x, loss)
        else:
            return x


class MiniBertForRegression(nn.Module):
    def __init__(self, configuration):
        super().__init__()
        self.minibert = MiniBert(configuration)
        self.configuration = configuration

        self._voc_size = len(configuration.vocabulary)
        self._embedding_dim = configuration.embedding_dim

        self.l1 = nn.Linear(self._embedding_dim,
                            configuration.first_layer_output_size, bias=False)
        self.l1_activation_fun = parse_activation_function(
            configuration.first_layer_activation_fun)
        self.l2 = nn.Linear(
            configuration.first_layer_output_size, configuration.output_size, bias=True)
        self.l2_activation_fun = parse_activation_function(
            configuration.second_layer_activation_fun)

    def forward(self, input):
        x = self.minibert(input)

        x = self.l1(x)
        x = self.l1_activation_fun(x)
        x = self.l2(x)
        x = self.l2_activation_fun(x)
        return x


class MiniBertForMLMAndRegression(nn.Module):
    def __init__(self, configuration):
        super().__init__()
        self.minibert = MiniBert(configuration)
        self.configuration = configuration

        self._voc_size = len(configuration.vocabulary)
        self._embedding_dim = configuration.embedding_dim

        self.mlm_l1 = nn.Linear(self._embedding_dim,
                                configuration.mlm_first_layer_output_size, bias=False)
        self.mlm_l2 = nn.Linear(
            configuration.mlm_first_layer_output_size, self._voc_size, bias=True)

        self.reg_l1 = nn.Linear(self._embedding_dim,
                                configuration.reg_first_layer_output_size, bias=False)
        self.reg_l2 = nn.Linear(
            configuration.reg_first_layer_output_size, configuration.reg_output_size, bias=True)

        self.mask_idx = configuration.mask_idx

        self.mlm_l1_activation_fun = parse_activation_function(
            configuration.mlm_first_layer_activation_fun)
        self.reg_l1_activation_fun = parse_activation_function(
            configuration.reg_first_layer_activation_fun)

        self.mlm_l2_activation_fun = parse_activation_function(
            configuration.mlm_second_layer_activation_fun)
        self.reg_l2_activation_fun = parse_activation_function(
            configuration.reg_second_layer_activation_fun)

        self._mask_prob = configuration.mask_prob
        self._keep_mask_prob = configuration.keep_mask_prob
        self._inv_corrupt_mask_prob = 1 - configuration.corrupt_mask_prob

    # task == 0 -> MLM
    # task == 1 -> Regression
    def forward(self, input, task):
        if task == 0:
            return self.forward_mlm(input)
        elif task == 1:
            return self.forward_reg(input)
        else:
            raise Exception(
                f"`task` parameter must be either 0 or 1, not {task}.")

    def forward_mlm(self, input):
        if self.training:
            masked_input = input.clone().detach()
            masked = torch.rand_like(
                input, dtype=torch.float) <= self._mask_prob

            masking_strategy = torch.rand_like(input, dtype=torch.float)
            masking = masked & (masking_strategy <=
                                self._keep_mask_prob)  # Keep masks
            corrupt = masked & (self._inv_corrupt_mask_prob <
                                masking_strategy)  # Corrupt masks

            replacements = torch.randint(
                self._voc_size, (torch.sum(corrupt), ), device=input.device)

            masked_input[masking] = self.mask_idx
            masked_input[corrupt] = replacements
            x = self.minibert(masked_input)
        else:
            x = self.minibert(input)

        x = self.mlm_l1(x)
        x = self.mlm_l1_activation_fun(x)
        x = self.mlm_l2(x)
        x = self.mlm_l2_activation_fun(x)

        if self.training:
            labels = input.clone().detach()
            labels[~masked] = -1

            loss_fn = nn.CrossEntropyLoss(ignore_index=-1)
            loss = loss_fn(x.view(-1, self._voc_size), labels.view(-1))
            return (x, loss)
        else:
            return x

    def forward_reg(self, input):
        x = self.minibert(input)
        x = self.reg_l1(x)
        x = self.reg_l1_activation_fun(x)
        x = self.reg_l2(x)
        x = self.reg_l2_activation_fun(x)
        return x
